"""CLI to test all API functions and models.

Execute with:
poetry run python3 -m myskoda.cli
"""

import asyncio
from logging import DEBUG, INFO

import asyncclick as click
import coloredlogs
from aiohttp import ClientSession
from asyncclick.core import Context
from termcolor import colored

from .event import Event, EventType, ServiceEventTopic
from .models.air_conditioning import AirConditioningState
from .models.charging import MaxChargeCurrent
from .models.common import (
    ActiveState,
    ChargerLockedState,
    ConnectionState,
    DoorLockedState,
    OnOffState,
    OpenState,
)
from .models.operation_request import OperationName, OperationStatus
from .myskoda import TRACE_CONFIG, MySkoda


@click.group()
@click.version_option()
@click.option("username", "--user", help="Username used for login.", required=True)
@click.option("password", "--password", help="Password used for login.", required=True)
@click.option("verbose", "--verbose", help="Enable verbose logging.", is_flag=True)
@click.pass_context
async def cli(ctx: Context, username: str, password: str, verbose: bool) -> None:
    """Interact with the MySkoda API."""
    coloredlogs.install(level=DEBUG if verbose else INFO)
    ctx.ensure_object(dict)
    ctx.obj["username"] = username
    ctx.obj["password"] = password

    trace_configs = []
    if verbose:
        trace_configs.append(TRACE_CONFIG)
    session = ClientSession(trace_configs=trace_configs)
    myskoda = MySkoda(session)
    await myskoda.connect(username, password)

    ctx.obj["myskoda"] = myskoda
    ctx.obj["session"] = session


@cli.result_callback()
@click.pass_context
async def disconnect(
    ctx: Context,
    result: None,  # noqa: ARG001
    username: str,  # noqa: ARG001
    password: str,  # noqa: ARG001
    verbose: bool,  # noqa: ARG001
) -> None:
    myskoda: MySkoda = ctx.obj["myskoda"]
    session: ClientSession = ctx.obj["session"]

    myskoda.disconnect()
    await session.close()


@cli.command()
@click.pass_context
async def auth(ctx: Context) -> None:
    """Extract the auth token."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    print(myskoda.get_auth_token())


@cli.command()
@click.pass_context
async def list_vehicles(ctx: Context) -> None:
    """Print a list of all vehicle identification numbers associated with the account."""
    myskoda: MySkoda = ctx.obj["myskoda"]

    print(f"{colored("vehicles:", "blue")}")
    for vehicle in await myskoda.list_vehicle_vins():
        print(f"- {vehicle}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def info(ctx: Context, vin: str) -> None:
    """Print info for the specified vin."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    info = await myskoda.get_info(vin)

    if info.specification.battery is not None:
        print(f"{colored("battery capacity:", "blue")} {info.specification.battery.capacity}kwh")
    print(f"{colored("power:", "blue")} {info.specification.engine.power}kw")
    print(f"{colored("engine:", "blue")} {info.specification.engine.type}")
    print(f"{colored("model:", "blue")} {info.specification.model}")
    print(f"{colored("model id:", "blue")} {info.specification.system_model_id}")
    print(f"{colored("model year:", "blue")} {info.specification.model_year}")
    print(f"{colored("title:", "blue")} {info.specification.title}")
    print(f"{colored("vin:", "blue")} {info.vin}")
    print(f"{colored("software:", "blue")} {info.software_version}")
    print(f"{colored("capabilities:", "blue")}")
    for capability in info.capabilities.capabilities:
        print(f"- {capability.id}")
        if len(capability.statuses) != 0:
            print("  status:")
            for status in capability.statuses:
                print(f"  - {status}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def status(ctx: Context, vin: str) -> None:
    """Print current status for the specified vin."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    status = await myskoda.get_status(vin)

    print(
        f"{colored("doors:", "blue")} {c_open(status.overall.doors)}, "
        f"{locked(status.overall.doors_locked)}"
    )
    print(f"{colored("bonnet:", "blue")} {c_open(status.detail.bonnet)}")
    print(f"{colored("trunk:", "blue")} {c_open(status.detail.trunk)}")
    print(f"{colored("windows:", "blue")} {c_open(status.overall.windows)}")
    print(f"{colored("lights:", "blue")} {on(status.overall.lights)}")
    print(f"{colored("last update:", "blue")} {status.car_captured_timestamp}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def air_conditioning(ctx: Context, vin: str) -> None:
    """Print current status about air conditioning."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    ac = await myskoda.get_air_conditioning(vin)

    print(f"{colored("window heating:", "blue")} {bool_state(ac.window_heating_enabled)}")
    print(f"{colored("window heating (front):", "blue")} {on(ac.window_heating_state.front)}")
    print(f"{colored("window heating (back):", "blue")} {on(ac.window_heating_state.rear)}")
    if ac.target_temperature is not None:
        print(
            f"{colored("target temperature:", "blue")} "
            f"{ac.target_temperature.temperature_value}°C"
        )
    print(f"{colored("steering wheel position:", "blue")} {ac.steering_wheel_position}")
    print(f"{colored("air conditioning:", "blue")} {ac_on(ac.state)}")
    print(f"{colored("state:", "blue")} {ac.state}")
    print(
        f"{colored("charger:", "blue")} {connected(ac.charger_connection_state)}, "
        f"{charger_locked(ac.charger_lock_state)}"
    )
    print(
        f"{colored("temperature reached:", "blue")} "
        f"{ac.estimated_date_time_to_reach_target_temperature}"
    )


@cli.command()
@click.argument("vin")
@click.pass_context
async def positions(ctx: Context, vin: str) -> None:
    """Print the vehicle's current position."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    positions = await myskoda.get_positions(vin)

    for position in positions.positions:
        print(f"- {colored("type:", "blue")} {position.type}")
        print(f"  {colored("latitude:", "blue")} {position.gps_coordinates.latitude}")
        print(f"  {colored("longitude:", "blue")} {position.gps_coordinates.longitude}")
        print(f"  {colored("address:", "blue")}")
        print(f"     {position.address.street} {position.address.house_number}")
        print(f"     {position.address.zip_code} {position.address.city}")
        print(f"     {position.address.country} ({position.address.country_code})")

    if positions.errors:
        print(f"{colored("Error:", "red")}")
    for error in positions.errors:
        print(f"- {colored("type:", "blue")} {error.type}")
        print(f"  {colored("description:", "blue")} {error.description}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def health(ctx: Context, vin: str) -> None:
    """Print the vehicle's mileage."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    health = await myskoda.get_health(vin)

    print(f"{colored("mileage:", "blue")} {health.mileage_in_km}km")
    print(f"{colored("last updated:", "blue")} {health.captured_at}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def charging(ctx: Context, vin: str) -> None:
    """Print the vehicle's current charging state."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    charging = await myskoda.get_charging(vin)

    if charging.status is not None:
        print(
            f"{colored("battery charge:", "blue")} "
            f"{charging.status.battery.state_of_charge_in_percent}%"
        )
        print(
            f"{colored("remaining distance:", "blue")} "
            f"{charging.status.battery.remaining_cruising_range_in_meters / 1000}km"
        )
        print(f"{colored("state:", "blue")} {charging.status.state}")
        print(f"{colored("charger type:", "blue")} {charging.status.charge_type}")
        print(
            f"{colored("charging rate:", "blue")} "
            f"{charging.status.charging_rate_in_kilometers_per_hour}km/h"
        )
        print(
            f"{colored("remaining time:", "blue")} "
            f"{charging.status.remaining_time_to_fully_charged_in_minutes}min"
        )
        print(f"{colored("charging power:", "blue")} {charging.status.charge_power_in_kw}kw")
    print(f"{colored("target:", "blue")} {charging.settings.target_state_of_charge_in_percent}%")
    print(
        f"{colored("battery care mode:", "blue")} "
        f"{active(charging.settings.charging_care_mode)}"
    )
    print(
        f"{colored("current:", "blue")} "
        f"{charge_current(charging.settings.max_charge_current_ac)}"
    )


@cli.command()
@click.argument("vin")
@click.pass_context
async def maintenance(ctx: Context, vin: str) -> None:
    """Print the vehicle's maintenance information."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    maintenance = await myskoda.get_maintenance(vin)

    if maintenance.maintenance_report:
        print(f"{colored("mileage:", "blue")} {maintenance.maintenance_report.mileage_in_km}km")
        print(
            f"{colored("inspection due in:", "blue")} "
            f"{maintenance.maintenance_report.inspection_due_in_days}d / "
            f"{maintenance.maintenance_report.inspection_due_in_km}km"
        )
        print(
            f"{colored("oil service due in:", "blue")} "
            f"{maintenance.maintenance_report.oil_service_due_in_days}d / "
            f"{maintenance.maintenance_report.oil_service_due_in_km}km"
        )
    if maintenance.preferred_service_partner:
        print(f"{colored("service partner:", "blue")} {maintenance.preferred_service_partner.name}")
        address = maintenance.preferred_service_partner.address
        print(f"     {address.street} {address.house_number}")
        print(f"     {address.zip_code} {address.city}")
        print(f"     {address.country} ({address.country_code})")
    if maintenance.predictive_maintenance:
        print(f"{colored("email:", "blue")} {maintenance.predictive_maintenance.setting.email}")
        print(
            f"{colored("phone:", "blue")} {maintenance.predictive_maintenance.setting.phone or '-'}"
        )


@cli.command()
@click.argument("vin")
@click.pass_context
async def driving_range(ctx: Context, vin: str) -> None:
    """Print the vehicle's estimated driving range information."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    driving_range = await myskoda.get_driving_range(vin)

    print(f"{colored("range:", "blue")} {driving_range.total_range_in_km}km")
    print(f"{colored("car type:", "blue")} {driving_range.car_type}")
    print(f"{colored("last update:", "blue")} {driving_range.car_captured_timestamp}")
    print(
        f"{colored("fuel level:", "blue")} "
        f"{driving_range.primary_engine_range.current_fuel_level_in_percent}%"
    )


@cli.command()
@click.pass_context
async def user(ctx: Context) -> None:
    """Print information about currently logged in user."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    user = await myskoda.get_user()

    print(f"{colored("id:", "blue")} {user.id}")
    print(f"{colored("name:", "blue")} {user.first_name} {user.last_name}")
    print(f"{colored("email:", "blue")} {user.email}")


@cli.command()
@click.argument("vin")
@click.pass_context
async def trip_statistics(ctx: Context, vin: str) -> None:
    """Print the last trip statics."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    stats = await myskoda.get_trip_statistics(vin)

    print(
        f"{colored("overall fuel consumption:", "blue")} "
        f"{stats.overall_average_fuel_consumption}l/100km"
    )
    print(f"{colored("mileage:", "blue")} {stats.overall_mileage_in_km}km")
    print(f"{colored("entries:", "blue")}")
    for entry in stats.detailed_statistics:
        print(f"- {colored("date:", "blue")} {entry.date}")
        print(
            f"  {colored("average fuel consumption:", "blue")} "
            f"{entry.average_fuel_consumption}l/100km"
        )
        print(f"  {colored("average speed:", "blue")} {entry.average_speed_in_kmph}km/h")
        print(f"  {colored("mileage:", "blue")} {entry.mileage_in_km}km")


@cli.command()
@click.argument("operation", type=click.Choice(OperationName))  # pyright: ignore [reportArgumentType]
@click.pass_context
async def wait_for_operation(ctx: Context, operation: OperationName) -> None:
    """Wait for the operation with the specified name to complete."""
    myskoda: MySkoda = ctx.obj["myskoda"]

    print(f"Waiting for an operation {colored(operation,"green")} to start and complete...")

    await myskoda.mqtt.wait_for_operation(operation)
    print("Completed.")


@cli.command()
@click.pass_context
async def subscribe(ctx: Context) -> None:
    """Connect to the MQTT broker and listen for messages."""
    myskoda: MySkoda = ctx.obj["myskoda"]

    def on_event(event: Event) -> None:
        print(f"{colored("- type:", "blue")} {event.type}")
        print(f"{colored("  vin:", "blue")} {event.vin}")
        if event.type == EventType.OPERATION:
            operation = event.operation
            print(f"  {colored("version:", "blue")} {operation.version}")
            print(f"  {colored("trace id:", "blue")} {operation.trace_id}")
            print(f"  {colored("request id:", "blue")} {operation.request_id}")
            print(f"  {colored("operation:", "blue")} {operation.operation}")
            print(f"  {colored("status:", "blue")} {operation.status}")
            if status == OperationStatus.ERROR:
                print(f"  {colored("error code:", "blue")} {operation.error_code}")
        elif event.type == EventType.SERVICE_EVENT:
            data = event.event.data
            print(f"  {colored("version:", "blue")} {event.event.version}")
            print(f"  {colored("trace id:", "blue")} {event.event.trace_id}")
            print(f"  {colored("timestamp:", "blue")} {event.event.timestamp}")
            print(f"  {colored("producer:", "blue")} {event.event.producer}")
            print(f"  {colored("name:", "blue")} {event.event.name}")
            print(f"  {colored("vin:", "blue")} {data.vin}")
            print(f"  {colored("user id:", "blue")} {data.user_id}")
            if event.topic == ServiceEventTopic.CHARGING:
                data = event.event.data
                print(f"  {colored("mode:", "blue")} {data.mode}")
                print(f"  {colored("state:", "blue")} {data.state}")
                print(f"  {colored("soc:", "blue")} {data.soc}%")
                print(f"  {colored("charged range:", "blue")} {data.charged_range}km")
                print(f"  {colored("time to finish:", "blue")} {data.time_to_finish}min")

    myskoda.subscribe(on_event)
    print(f"{colored("Listening for Mqtt events:", "blue")}")
    await asyncio.Event().wait()


@cli.command()
@click.option("temperature", "--temperature", type=float, required=True)
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
async def start_air_conditioning(
    ctx: Context,
    temperature: float,
    timeout: float,  # noqa: ASYNC109
    vin: str,
) -> None:
    """Start the air conditioning with the provided target temperature in °C."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.start_air_conditioning(vin, temperature)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
async def stop_air_conditioning(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Stop the air conditioning."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.stop_air_conditioning(vin)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("temperature", "--temperature", type=float, required=True)
@click.pass_context
async def set_target_temperature(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    temperature: float,
) -> None:
    """Set the air conditioning's target temperature in °C."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_target_temperature(vin, temperature)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
async def start_window_heating(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Start heating both the front and rear window."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.start_window_heating(vin)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
async def stop_window_heating(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Stop heating both the front and rear window."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.stop_window_heating(vin)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("limit", "--limit", type=float, required=True)
@click.pass_context
async def set_charge_limit(ctx: Context, timeout: float, vin: str, limit: int) -> None:  # noqa: ASYNC109
    """Set the maximum charge limit in percent."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_charge_limit(vin, limit)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
async def set_battery_care_mode(ctx: Context, timeout: float, vin: str, enabled: bool) -> None:  # noqa: ASYNC109
    """Enable or disable the battery care mode."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_battery_care_mode(vin, enabled)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
async def set_reduced_current_limit(ctx: Context, timeout: float, vin: str, enabled: bool) -> None:  # noqa: ASYNC109
    """Enable reducing the current limit by which the car is charged."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_reduced_current_limit(vin, enabled)


@cli.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
async def wakeup(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Wake the vehicle up. Can be called maximum three times a day."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.wakeup(vin)


def c_open(cond: OpenState) -> str:
    return colored("open", "red") if cond == OpenState.OPEN else colored("closed", "green")


def locked(cond: DoorLockedState) -> str:
    return (
        colored("locked", "green") if cond == DoorLockedState.LOCKED else colored("unlocked", "red")
    )


def on(cond: OnOffState) -> str:
    return colored("on", "green") if cond == OnOffState.ON else colored("off", "red")


def ac_on(cond: AirConditioningState) -> str:
    return (
        colored("on", "green")
        if cond in (AirConditioningState.ON, AirConditioningState.HEATING)
        else colored("off", "red")
    )


def connected(cond: ConnectionState) -> str:
    return (
        colored("connected", "green")
        if cond == ConnectionState.CONNECTED
        else colored("disconnected", "red")
    )


def active(cond: ActiveState) -> str:
    return (
        colored("active", "green") if cond == ActiveState.ACTIVATED else colored("inactive", "red")
    )


def charge_current(cond: MaxChargeCurrent) -> str:
    return (
        colored("maximum", "green")
        if cond == MaxChargeCurrent.MAXIMUM
        else colored("reduced", "red")
    )


def bool_state(cond: bool) -> str:
    return colored("true", "green") if cond else colored("false", "red")


def charger_locked(cond: ChargerLockedState) -> str:
    return (
        colored("locked", "green")
        if cond == ChargerLockedState.LOCKED
        else colored("unlocked", "red")
    )


if __name__ == "__main__":
    cli()



"""Contains API representation for the MySkoda REST API."""

import logging
from collections.abc import Callable

from aiohttp import ClientSession

from .auth.authorization import Authorization
from .const import BASE_URL_SKODA
from .models.air_conditioning import AirConditioning
from .models.charging import ChargeMode, Charging
from .models.driving_range import DrivingRange
from .models.health import Health
from .models.info import Info
from .models.maintenance import Maintenance
from .models.position import Positions, PositionType
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User

_LOGGER = logging.getLogger(__name__)


class RestApi:
    """API hub class that can perform all calls to the MySkoda API."""

    session: ClientSession
    authorization: Authorization

    def __init__(self, session: ClientSession, authorization: Authorization) -> None:  # noqa: D107
        self.session = session
        self.authorization = authorization

    async def _make_get_request[T](self, url: str, deserialize: Callable[[str], T]) -> T:
        async with self.session.get(
            f"{BASE_URL_SKODA}/api{url}",
            headers=await self._headers(),
        ) as response:
            response_text = await response.text()
            try:
                data = deserialize(response_text)
            except Exception:
                _LOGGER.exception(
                    "Failed to load data from url %s. Return value was '%s'", url, response_text
                )
                raise
            else:
                return data

    async def get_info(self, vin: str) -> Info:
        """Retrieve the basic vehicle information for the specified vehicle."""
        return await self._make_get_request(
            f"/v2/garage/vehicles/{vin}?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4",
            Info.from_json,
        )

    async def get_charging(self, vin: str) -> Charging:
        """Retrieve information related to charging for the specified vehicle."""
        return await self._make_get_request(f"/v1/charging/{vin}", Charging.from_json)

    async def get_status(self, vin: str) -> Status:
        """Retrieve the current status for the specified vehicle."""
        return await self._make_get_request(f"/v2/vehicle-status/{vin}", Status.from_json)

    async def get_air_conditioning(self, vin: str) -> AirConditioning:
        """Retrieve the current air conditioning status for the specified vehicle."""
        return await self._make_get_request(
            f"/v2/air-conditioning/{vin}", AirConditioning.from_json
        )

    async def get_positions(self, vin: str) -> Positions:
        """Retrieve the current position for the specified vehicle."""
        return await self._make_get_request(f"/v1/maps/positions?vin={vin}", Positions.from_json)

    async def get_driving_range(self, vin: str) -> DrivingRange:
        """Retrieve estimated driving range for combustion vehicles."""
        return await self._make_get_request(
            f"/v2/vehicle-status/{vin}/driving-range", DrivingRange.from_json
        )

    async def get_trip_statistics(self, vin: str) -> TripStatistics:
        """Retrieve statistics about past trips."""
        return await self._make_get_request(
            f"/v1/trip-statistics/{vin}?offsetType=week&offset=0&timezone=Europe%2FBerlin",
            TripStatistics.from_json,
        )

    async def get_maintenance(self, vin: str) -> Maintenance:
        """Retrieve maintenance report."""
        return await self._make_get_request(
            f"/v3/vehicle-maintenance/vehicles/{vin}", Maintenance.from_json
        )

    async def get_health(self, vin: str) -> Health:
        """Retrieve health information for the specified vehicle."""
        return await self._make_get_request(
            f"/v1/vehicle-health-report/warning-lights/{vin}", Health.from_json
        )

    async def get_user(self) -> User:
        """Retrieve user information about logged in user."""
        return await self._make_get_request("/v1/users", User.from_json)

    async def list_vehicles(self) -> list[str]:
        """List all vehicles by their vins."""
        async with self.session.get(
            f"{BASE_URL_SKODA}/api/v2/garage?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4",
            headers=await self._headers(),
        ) as response:
            json = await response.json()
            return [vehicle["vin"] for vehicle in json["vehicles"]]

    async def _headers(self) -> dict[str, str]:
        return {"authorization": f"Bearer {await self.authorization.get_access_token()}"}

    async def stop_air_conditioning(self, vin: str) -> None:
        """Stop the air conditioning."""
        _LOGGER.debug("Stopping air conditioning for vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v2/air-conditioning/{vin}/stop",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def start_air_conditioning(self, vin: str, temperature: float) -> None:
        """Start the air conditioning."""
        _LOGGER.debug(
            "Starting air conditioning for vehicle %s with temperature %s",
            vin,
            str(temperature),
        )
        json_data = {
            "heaterSource": "ELECTRIC",
            "targetTemperature": {
                "temperatureValue": str(temperature),
                "unitInCar": "CELSIUS",
            },
        }
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v2/air-conditioning/{vin}/start",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    async def set_target_temperature(self, vin: str, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        _LOGGER.debug("Setting target temperature for vehicle %s to %s", vin, str(temperature))
        json_data = {"temperatureValue": str(temperature), "unitInCar": "CELSIUS"}
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v2/air-conditioning/{vin}/settings/target-temperature",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    async def start_window_heating(self, vin: str) -> None:
        """Start heating both the front and rear window."""
        _LOGGER.debug("Starting window heating for vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v2/air-conditioning/{vin}/start-window-heating",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def stop_window_heating(self, vin: str) -> None:
        """Stop heating both the front and rear window."""
        _LOGGER.debug("Stopping window heating for vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v2/air-conditioning/{vin}/stop-window-heating",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def set_charge_limit(self, vin: str, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        _LOGGER.debug("Setting charge limit for vehicle %s to %d", vin, limit)
        json_data = {"targetSOCInPercent": limit}
        async with self.session.put(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/set-charge-limit",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_battery_care_mode(self, vin: str, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        _LOGGER.debug("Setting battery care mode for vehicle %s to %r", vin, enabled)
        json_data = {"chargingCareMode": "ACTIVATED" if enabled else "DEACTIVATED"}
        async with self.session.put(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/set-care-mode",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_reduced_current_limit(self, vin: str, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        _LOGGER.debug("Setting reduced charging for vehicle %s to %r", vin, reduced)
        json_data = {"chargingCurrent": "REDUCED" if reduced else "MAXIMUM"}
        async with self.session.put(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/set-charging-current",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    async def start_charging(self, vin: str) -> None:
        """Start charging the car."""
        _LOGGER.debug("Starting charging for vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/start",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def stop_charging(self, vin: str) -> None:
        """Stop charging the car."""
        _LOGGER.debug("Stopping charging of vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/stop",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def wakeup(self, vin: str) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Waking up vehicle %s", vin)
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/vehicle-wakeup/{vin}?applyRequestLimiter=true",
            headers=await self._headers(),
        ) as response:
            await response.text()

    async def set_charge_mode(self, vin: str, mode: ChargeMode) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Changing charging mode of vehicle %s to %s", vin, mode)
        json_data = {"chargeMode": mode.value}
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/charging/{vin}/set-charge-mode",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()

    # TODO @dvx76: Maybe refactor for FBT001
    async def honk_flash(
        self,
        vin: str,
        honk: bool = False,  # noqa:FBT002
    ) -> None:
        """Honk and/or flash."""
        positions = await self.get_positions(vin)
        position = next(pos for pos in positions.positions if pos.type == PositionType.VEHICLE)
        json_data = {
            "mode": "HONK_AND_FLASH" if honk else "FLASH",
            "vehiclePosition": {
                "lat": position.gps_coordinates.latitude,
                "lng": position.gps_coordinates.longitude,
            },
        }
        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/vehicle-access/{vin}/honk-and-flash",
            headers=await self._headers(),
            json=json_data,
        ) as response:
            await response.text()



"""Models for responses of api/v1/charging endpoint."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .common import ActiveState, EnabledState


class ChargingErrorType(StrEnum):
    CARE_MODE_IS_NOT_AVAILABLE = "CARE_MODE_IS_NOT_AVAILABLE"
    AUTO_UNLOCK_IS_NOT_AVAILABLE = "AUTO_UNLOCK_IS_NOT_AVAILABLE"
    MAX_CHARGE_CURRENT_IS_NOT_AVAILABLE = "MAX_CHARGE_CURRENT_IS_NOT_AVAILABLE"
    CHARGE_LIMIT_IS_NOT_AVAILABLE = "CHARGE_LIMIT_IS_NOT_AVAILABLE"
    STATUS_OF_CHARGING_NOT_AVAILABLE = "STATUS_OF_CHARGING_NOT_AVAILABLE"
    STATUS_OF_CONNECTION_NOT_AVAILABLE = "STATUS_OF_CONNECTION_NOT_AVAILABLE"


@dataclass
class ChargingError(DataClassORJSONMixin):
    type: ChargingErrorType
    description: str


class ChargeMode(StrEnum):
    HOME_STORAGE_CHARGING = "HOME_STORAGE_CHARGING"
    IMMEDIATE_DISCHARGING = "IMMEDIATE_DISCHARGING"
    ONLY_OWN_CURRENT = "ONLY_OWN_CURRENT"
    PREFERRED_CHARGING_TIMES = "PREFERRED_CHARGING_TIMES"
    TIMER_CHARGING_WITH_CLIMATISATION = "TIMER_CHARGING_WITH_CLIMATISATION"
    TIMER = "TIMER"
    MANUAL = "MANUAL"
    OFF = "OFF"


class MaxChargeCurrent(StrEnum):
    MAXIMUM = "MAXIMUM"
    REDUCED = "REDUCED"


class ChargingState(StrEnum):
    READY_FOR_CHARGING = "READY_FOR_CHARGING"
    CONNECT_CABLE = "CONNECT_CABLE"
    CONSERVING = "CONSERVING"
    CHARGING = "CHARGING"


class ChargeType(StrEnum):
    AC = "AC"
    DC = "DC"


class PlugUnlockMode(StrEnum):
    PERMANENT = "PERMANENT"
    ON = "ON"
    OFF = "OFF"


@dataclass
class Settings(DataClassORJSONMixin):
    available_charge_modes: list[ChargeMode] = field(
        metadata=field_options(alias="availableChargeModes")
    )
    battery_support: EnabledState = field(metadata=field_options(alias="batterySupport"))
    charging_care_mode: ActiveState = field(metadata=field_options(alias="chargingCareMode"))
    max_charge_current_ac: MaxChargeCurrent = field(
        metadata=field_options(alias="maxChargeCurrentAc")
    )
    preferred_charge_mode: ChargeMode = field(metadata=field_options(alias="preferredChargeMode"))
    target_state_of_charge_in_percent: int = field(
        metadata=field_options(alias="targetStateOfChargeInPercent")
    )
    auto_unlock_plug_when_charged: PlugUnlockMode = field(
        metadata=field_options(alias="autoUnlockPlugWhenCharged")
    )


@dataclass
class Battery(DataClassORJSONMixin):
    remaining_cruising_range_in_meters: int = field(
        metadata=field_options(alias="remainingCruisingRangeInMeters")
    )
    state_of_charge_in_percent: int = field(metadata=field_options(alias="stateOfChargeInPercent"))


@dataclass
class ChargingStatus(DataClassORJSONMixin):
    battery: Battery
    state: ChargingState
    charging_rate_in_kilometers_per_hour: float = field(
        metadata=field_options(alias="chargingRateInKilometersPerHour")
    )
    charge_power_in_kw: float | None = field(
        default=None, metadata=field_options(alias="chargePowerInKw")
    )
    charge_type: ChargeType | None = field(default=None, metadata=field_options(alias="chargeType"))
    remaining_time_to_fully_charged_in_minutes: int | None = field(
        default=None, metadata=field_options(alias="remainingTimeToFullyChargedInMinutes")
    )


@dataclass
class Charging(DataClassORJSONMixin):
    """Information related to charging an EV."""

    errors: list[ChargingError]
    settings: Settings
    status: ChargingStatus | None
    car_captured_timestamp: datetime = field(metadata=field_options(alias="carCapturedTimestamp"))
    is_vehicle_in_saved_location: bool = field(
        metadata=field_options(alias="isVehicleInSavedLocation")
    )



"""Main entry point for the MySkoda library.

This class provides all methods to operate on the API and MQTT broker.
"""

import logging
from asyncio import gather
from collections.abc import Awaitable, Callable
from ssl import SSLContext
from types import SimpleNamespace

from aiohttp import ClientSession, TraceConfig, TraceRequestEndParams

from .auth.authorization import Authorization
from .event import Event
from .models.air_conditioning import AirConditioning
from .models.charging import Charging
from .models.driving_range import DrivingRange
from .models.health import Health
from .models.info import CapabilityId, Info
from .models.maintenance import Maintenance
from .models.operation_request import OperationName
from .models.position import Positions
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User
from .mqtt import Mqtt
from .rest_api import RestApi
from .vehicle import Vehicle

_LOGGER = logging.getLogger(__name__)


async def trace_response(
    _session: ClientSession,
    _trace_config_ctx: SimpleNamespace,
    params: TraceRequestEndParams,
) -> None:
    """Log response details. Used in aiohttp.TraceConfig."""
    _LOGGER.debug(
        f"Trace: {params.method} {str(params.url)[:60]} - "
        f"response: {params.response.status} ({params.response.content_length} bytes) "
        f"{(await params.response.text())[:5000]}"
    )


TRACE_CONFIG = TraceConfig()
TRACE_CONFIG.on_request_end.append(trace_response)


class MySkoda:
    session: ClientSession
    rest_api: RestApi
    mqtt: Mqtt
    authorization: Authorization

    def __init__(  # noqa: D107
        self, session: ClientSession, ssl_context: SSLContext | None = None
    ) -> None:
        self.session = session
        self.authorization = Authorization(session)
        self.rest_api = RestApi(self.session, self.authorization)
        self.mqtt = Mqtt(self.authorization, self.rest_api, ssl_context=ssl_context)

    async def connect(self, email: str, password: str) -> None:
        """Authenticate on the rest api and connect to the MQTT broker."""
        await self.authorization.authorize(email, password)
        _LOGGER.info("IDK Authorization was successful.")

        await self.mqtt.connect()
        _LOGGER.debug("Myskoda ready.")

    def subscribe(self, callback: Callable[[Event], None | Awaitable[None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        self.mqtt.subscribe(callback)

    def disconnect(self) -> None:
        """Disconnect from the MQTT broker."""
        self.mqtt.disconnect()

    async def stop_charging(self, vin: str) -> None:
        """Stop the car from charging."""
        future = self.mqtt.wait_for_operation(OperationName.STOP_CHARGING)
        await self.rest_api.stop_charging(vin)
        await future

    async def start_charging(self, vin: str) -> None:
        """Start charging the car."""
        future = self.mqtt.wait_for_operation(OperationName.START_CHARGING)
        await self.rest_api.start_charging(vin)
        await future

    async def honk_flash(self, vin: str, honk: bool = False) -> None:  # noqa: FBT002
        """Honk and/or flash."""
        future = self.mqtt.wait_for_operation(OperationName.START_HONK)
        await self.rest_api.honk_flash(vin, honk)
        await future

    async def wakeup(self, vin: str) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        future = self.mqtt.wait_for_operation(OperationName.WAKEUP)
        await self.rest_api.honk_flash(vin)
        await future

    async def set_reduced_current_limit(self, vin: str, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        future = self.mqtt.wait_for_operation(OperationName.UPDATE_CHARGING_CURRENT)
        await self.rest_api.set_reduced_current_limit(vin, reduced=reduced)
        await future

    async def set_battery_care_mode(self, vin: str, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        future = self.mqtt.wait_for_operation(OperationName.UPDATE_CARE_MODE)
        await self.rest_api.set_battery_care_mode(vin, enabled)
        await future

    async def set_charge_limit(self, vin: str, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        future = self.mqtt.wait_for_operation(OperationName.UPDATE_CHARGE_LIMIT)
        await self.rest_api.set_charge_limit(vin, limit)
        await future

    async def stop_window_heating(self, vin: str) -> None:
        """Stop heating both the front and rear window."""
        future = self.mqtt.wait_for_operation(OperationName.STOP_WINDOW_HEATING)
        await self.rest_api.stop_window_heating(vin)
        await future

    async def start_window_heating(self, vin: str) -> None:
        """Start heating both the front and rear window."""
        future = self.mqtt.wait_for_operation(OperationName.START_WINDOW_HEATING)
        await self.rest_api.start_window_heating(vin)
        await future

    async def set_target_temperature(self, vin: str, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        future = self.mqtt.wait_for_operation(OperationName.SET_AIR_CONDITIONING_TARGET_TEMPERATURE)
        await self.rest_api.set_target_temperature(vin, temperature)
        await future

    async def start_air_conditioning(self, vin: str, temperature: float) -> None:
        """Start the air conditioning with the provided target temperature in °C."""
        future = self.mqtt.wait_for_operation(OperationName.START_AIR_CONDITIONING)
        await self.rest_api.start_air_conditioning(vin, temperature)
        await future

    async def stop_air_conditioning(self, vin: str) -> None:
        """Stop the air conditioning."""
        future = self.mqtt.wait_for_operation(OperationName.STOP_AIR_CONDITIONING)
        await self.rest_api.stop_air_conditioning(vin)
        await future

    async def get_auth_token(self) -> str:
        """Retrieve the main access token for the IDK session."""
        return await self.rest_api.authorization.get_access_token()

    async def get_info(self, vin: str) -> Info:
        """Retrieve the basic vehicle information for the specified vehicle."""
        return await self.rest_api.get_info(vin)

    async def get_charging(self, vin: str) -> Charging:
        """Retrieve information related to charging for the specified vehicle."""
        return await self.rest_api.get_charging(vin)

    async def get_status(self, vin: str) -> Status:
        """Retrieve the current status for the specified vehicle."""
        return await self.rest_api.get_status(vin)

    async def get_air_conditioning(self, vin: str) -> AirConditioning:
        """Retrieve the current air conditioning status for the specified vehicle."""
        return await self.rest_api.get_air_conditioning(vin)

    async def get_positions(self, vin: str) -> Positions:
        """Retrieve the current position for the specified vehicle."""
        return await self.rest_api.get_positions(vin)

    async def get_driving_range(self, vin: str) -> DrivingRange:
        """Retrieve estimated driving range for combustion vehicles."""
        return await self.rest_api.get_driving_range(vin)

    async def get_trip_statistics(self, vin: str) -> TripStatistics:
        """Retrieve statistics about past trips."""
        return await self.rest_api.get_trip_statistics(vin)

    async def get_maintenance(self, vin: str) -> Maintenance:
        """Retrieve maintenance report."""
        return await self.rest_api.get_maintenance(vin)

    async def get_health(self, vin: str) -> Health:
        """Retrieve health information for the specified vehicle."""
        return await self.rest_api.get_health(vin)

    async def get_user(self) -> User:
        """Retrieve user information about logged in user."""
        return await self.rest_api.get_user()

    async def list_vehicle_vins(self) -> list[str]:
        """List all vehicles by their vins."""
        return await self.rest_api.list_vehicles()

    async def get_vehicle(self, vin: str) -> Vehicle:
        """Load a full vehicle based on its capabilities."""
        info = await self.get_info(vin)
        maintenance = await self.get_maintenance(vin)

        vehicle = Vehicle(info, maintenance)

        if info.is_capability_available(CapabilityId.STATE):
            vehicle.status = await self.get_status(vin)
            vehicle.driving_range = await self.get_driving_range(vin)

        if info.is_capability_available(CapabilityId.AIR_CONDITIONING):
            vehicle.air_conditioning = await self.get_air_conditioning(vin)

        if info.is_capability_available(CapabilityId.PARKING_POSITION):
            vehicle.positions = await self.get_positions(vin)

        if info.is_capability_available(CapabilityId.TRIP_STATISTICS):
            vehicle.trip_statistics = await self.get_trip_statistics(vin)

        if info.is_capability_available(CapabilityId.CHARGING):
            vehicle.charging = await self.get_charging(vin)

        if info.is_capability_available(CapabilityId.VEHICLE_HEALTH_INSPECTION):
            vehicle.health = await self.get_health(vin)

        return vehicle

    async def get_all_vehicles(self) -> list[Vehicle]:
        """Load all vehicles based on their capabilities."""
        vins = await self.list_vehicle_vins()
        return await gather(*(self.get_vehicle(vin) for vin in vins))



"""Models for responses of api/v2/vehicle-status/{vin}."""

from dataclasses import dataclass, field
from datetime import datetime

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from myskoda.models.common import DoorLockedState, OnOffState, OpenState


@dataclass
class Detail(DataClassORJSONMixin):
    bonnet: OpenState
    sunroof: OpenState
    trunk: OpenState


@dataclass
class Overall(DataClassORJSONMixin):
    doors: OpenState
    doors_locked: DoorLockedState = field(metadata=field_options(alias="doorsLocked"))
    lights: OnOffState
    locked: DoorLockedState
    windows: OpenState


@dataclass
class Status(DataClassORJSONMixin):
    """Current status information for a vehicle."""

    detail: Detail
    overall: Overall
    car_captured_timestamp: datetime | None = field(
        default=None, metadata=field_options(alias="carCapturedTimestamp")
    )



"""Models related to service events from the MQTT broker."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from typing import Generic, TypeVar

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .charging import ChargeMode, ChargingState


class ServiceEventName(StrEnum):
    CHANGE_SOC = "change-soc"
    CHANGE_ACCESS = "change-access"
    CHANGE_LIGHTS = "change-lights"
    CLIMATISATION_COMPLETED = "climatisation-completed"
    CHANGE_REMAINING_TIME = "change-remaining-time"
    CHANGE_CHARGE_MODE = "change-charge-mode"


@dataclass
class ServiceEventData(DataClassORJSONMixin):
    user_id: str = field(metadata=field_options(alias="userId"))
    vin: str


T = TypeVar("T", bound=ServiceEventData)


@dataclass
class ServiceEvent(Generic[T], DataClassORJSONMixin):
    version: int
    producer: str
    name: ServiceEventName
    data: T
    trace_id: str = field(metadata=field_options(alias="traceId"))
    timestamp: datetime | None = field(default=None, metadata=field_options(alias="requestId"))


def _deserialize_mode(value: str) -> ChargeMode:  # noqa: PLR0911
    match value:
        case "homeStorageCharging":
            return ChargeMode.HOME_STORAGE_CHARGING
        case "immediateDischarging":
            return ChargeMode.IMMEDIATE_DISCHARGING
        case "onlyOwnCurrent":
            return ChargeMode.ONLY_OWN_CURRENT
        case "preferredChargingTimes":
            return ChargeMode.PREFERRED_CHARGING_TIMES
        case "timerChargingWithClimatisation":
            return ChargeMode.TIMER_CHARGING_WITH_CLIMATISATION
        case "timer":
            return ChargeMode.TIMER
        case "manual":
            return ChargeMode.MANUAL
        case "off":
            return ChargeMode.OFF
        case _:
            raise UnexpectedChargeModeError


def _deserialize_charging_state(value: str) -> ChargingState:
    match value:
        case "charging":
            return ChargingState.CHARGING
        case "chargePurposeReachedAndNotConservationCharging":
            return ChargingState.READY_FOR_CHARGING
        case "notReadyForCharging":
            return ChargingState.CONNECT_CABLE
        case "readyForCharging":
            return ChargingState.READY_FOR_CHARGING
        case "conserving":
            return ChargingState.CONSERVING
        case _:
            raise UnexpectedChargingStateError


@dataclass
class ServiceEventChargingData(ServiceEventData):
    mode: ChargeMode
    state: ChargingState
    soc: int
    charged_range: int = field(
        metadata=field_options(alias="chargedRange", deserialize=_deserialize_charging_state)
    )
    time_to_finish: int | None = field(
        default=None,
        metadata=field_options(alias="timeToFinish", deserialize=_deserialize_mode),
    )


@dataclass
class ServiceEventCharging(ServiceEvent):
    data: ServiceEventChargingData


class UnexpectedChargeModeError(Exception):
    pass


class UnexpectedChargingStateError(Exception):
    pass



"""Events emitted by MySkoda."""

from datetime import UTC, datetime
from enum import StrEnum
from typing import Literal

from .models.operation_request import OperationRequest
from .models.service_event import ServiceEvent, ServiceEventCharging


class ServiceEventTopic(StrEnum):
    ACCESS = "ACCESS"
    AIR_CONDITIONING = "AIR_CONDITIONING"
    CHARGING = "CHARGING"
    LIGHTS = "LIGHTS"


class AccountEventTopic(StrEnum):
    ACCOUNT_PRIVACY = "ACCOUNT_PRIVACY"


class EventType(StrEnum):
    ACCOUNT_EVENT = "account-event"
    OPERATION = "operation-request"
    SERVICE_EVENT = "service-event"


class BaseEvent:
    vin: str
    user_id: str
    timestamp: datetime

    def __init__(self, vin: str, user_id: str) -> None:  # noqa: D107
        self.vin = vin
        self.user_id = user_id
        self.timestamp = datetime.now(tz=UTC)


class EventOperation(BaseEvent):
    type: Literal[EventType.OPERATION] = EventType.OPERATION
    operation: OperationRequest

    def __init__(self, vin: str, user_id: str, operation: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)
        self.operation = OperationRequest.from_dict(operation)


class EventAirConditioning(BaseEvent):
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.AIR_CONDITIONING] = ServiceEventTopic.AIR_CONDITIONING
    event: ServiceEvent

    def __init__(self, vin: str, user_id: str, payload: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)
        self.event = ServiceEvent.from_dict(payload)


class EventCharging(BaseEvent):
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.CHARGING] = ServiceEventTopic.CHARGING
    event: ServiceEventCharging

    def __init__(self, vin: str, user_id: str, payload: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)
        self.event = ServiceEventCharging.from_dict(payload)


class EventAccess(BaseEvent):
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.ACCESS] = ServiceEventTopic.ACCESS
    event: ServiceEvent

    def __init__(self, vin: str, user_id: str, payload: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)
        self.event = ServiceEvent.from_dict(payload)


class EventLights(BaseEvent):
    type: Literal[EventType.SERVICE_EVENT] = EventType.SERVICE_EVENT
    topic: Literal[ServiceEventTopic.LIGHTS] = ServiceEventTopic.LIGHTS
    event: ServiceEvent

    def __init__(self, vin: str, user_id: str, payload: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)
        self.event = ServiceEvent.from_dict(payload)


class EventAccountPrivacy(BaseEvent):
    type: Literal[EventType.ACCOUNT_EVENT] = EventType.ACCOUNT_EVENT
    topic: Literal[AccountEventTopic.ACCOUNT_PRIVACY] = AccountEventTopic.ACCOUNT_PRIVACY

    def __init__(self, vin: str, user_id: str, _payload: dict) -> None:  # noqa: D107
        super().__init__(vin, user_id)


Event = (
    EventAccountPrivacy
    | EventOperation
    | EventAccess
    | EventAirConditioning
    | EventCharging
    | EventLights
)



"""Parse CSRF information from the website."""

import re
from dataclasses import dataclass, field
from html.parser import HTMLParser

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin

json_object = re.compile(r"window\._IDK\s=\s((?:\n|.)*?)$")


@dataclass
class TemplateModel(DataClassORJSONMixin):
    hmac: str
    relay_state: str = field(metadata=field_options(alias="relayState"))


@dataclass
class CSRFState(DataClassYAMLMixin):
    csrf: str = field(metadata=field_options(alias="csrf_token"))
    template_model: TemplateModel = field(metadata=field_options(alias="templateModel"))


class CSRFParser(HTMLParser):
    """Information such as the CSRF or the hmac will be available in the HTML.

    This will parse the information from a `<script>` tag in the HTML.
    """

    _is_script = False
    csrf_state: None | CSRFState = None

    def handle_starttag(
        self,
        tag: str,
        attrs: list[tuple[str, str | None]],  # noqa: ARG002
    ) -> None:
        """Determine whether a script tag has been entered."""
        if tag != "script":
            return
        self._is_script = True

    def handle_endtag(self, tag: str) -> None:
        """Determine whether a script tag has been left."""
        if tag != "script":
            return
        self._is_script = False

    def handle_data(self, data: str) -> None:
        """Parse the contents of a script tag to extract csrf information."""
        if not self._is_script:
            return

        result = json_object.search(data)
        if result is None:
            return

        result = result.group(1)
        # Load the info using YAML, since the syntax used in the script is YAML compatible,
        # but not JSON compatible (missing quotes around field names, trailing commas).
        self.csrf_state = CSRFState.from_yaml(result)



"""Models for responses of api/v3/vehicle-maintenance/vehicles/{vin} endpoint."""

from dataclasses import dataclass, field
from datetime import datetime, time
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .common import Address, Coordinates, Weekday


@dataclass
class MaintenanceReport(DataClassORJSONMixin):
    captured_at: datetime = field(metadata=field_options(alias="capturedAt"))
    inspection_due_in_days: int = field(metadata=field_options(alias="inspectionDueInDays"))
    mileage_in_km: int | None = field(default=None, metadata=field_options(alias="mileageInKm"))
    inspection_due_in_km: int | None = field(
        default=None, metadata=field_options(alias="inspectionDueInKm")
    )
    oil_service_due_in_days: int | None = field(
        default=None, metadata=field_options(alias="oilServiceDueInDays")
    )
    oil_service_due_in_km: int | None = field(
        default=None, metadata=field_options(alias="oilServiceDueInKm")
    )


@dataclass
class Contact(DataClassORJSONMixin):
    email: str | None = field(default=None)
    phone: str | None = field(default=None)
    url: str | None = field(default=None)


@dataclass
class TimeRange(DataClassORJSONMixin):
    start: time = field(metadata=field_options(alias="from"))
    end: time = field(metadata=field_options(alias="to"))


@dataclass
class OpeningHoursPeriod(DataClassORJSONMixin):
    opening_times: list[TimeRange] = field(metadata=field_options(alias="openingTimes"))
    period_end: Weekday = field(metadata=field_options(alias="periodEnd"))
    period_start: Weekday = field(metadata=field_options(alias="periodStart"))


class CommunicationChannel(StrEnum):
    email = "EMAIL"


@dataclass
class PredictiveMaintenanceSettings(DataClassORJSONMixin):
    email: str
    service_activated: bool = field(metadata=field_options(alias="serviceActivated"))
    phone: str | None = field(default=None)
    preferred_channel: CommunicationChannel | None = field(
        default=None,
        metadata=field_options(alias="preferredChannel"),
    )


@dataclass
class PredictiveMaintenance(DataClassORJSONMixin):
    setting: PredictiveMaintenanceSettings


@dataclass
class ServicePartner(DataClassORJSONMixin):
    address: Address
    brand: str
    contact: Contact
    id: str
    location: Coordinates
    name: str
    opening_hours: list[OpeningHoursPeriod] = field(metadata=field_options(alias="openingHours"))
    partner_number: str = field(metadata=field_options(alias="partnerNumber"))


@dataclass
class Maintenance(DataClassORJSONMixin):
    maintenance_report: MaintenanceReport | None = field(
        default=None, metadata=field_options(alias="maintenanceReport")
    )
    predictive_maintenance: PredictiveMaintenance | None = field(
        default=None, metadata=field_options(alias="predictiveMaintenance")
    )
    preferred_service_partner: ServicePartner | None = field(
        default=None, metadata=field_options(alias="preferredServicePartner")
    )



"""Models for responses of api/v2/vehicle-status/{vin}."""

from dataclasses import dataclass, field
from datetime import date
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


class VehicleType(StrEnum):
    FUEL = "FUEL"


@dataclass
class StatisticsEntry(DataClassORJSONMixin):
    date: date
    average_fuel_consumption: float | None = field(
        default=None, metadata=field_options(alias="averageFuelConsumption")
    )
    average_speed_in_kmph: int | None = field(
        default=None, metadata=field_options(alias="averageSpeedInKmph")
    )
    mileage_in_km: int | None = field(default=None, metadata=field_options(alias="mileageInKm"))
    travel_time_in_min: int | None = field(
        default=None, metadata=field_options(alias="travelTimeInMin")
    )
    trip_ids: list[int] | None = field(default=None, metadata=field_options(alias="tripIds"))


@dataclass
class TripStatistics(DataClassORJSONMixin):
    overall_average_fuel_consumption: float = field(
        metadata=field_options(alias="overallAverageFuelConsumption")
    )
    overall_average_mileage_in_km: int = field(
        metadata=field_options(alias="overallAverageMileageInKm")
    )
    overall_average_speed_in_kmph: int = field(
        metadata=field_options(alias="overallAverageSpeedInKmph")
    )
    overall_average_travel_time_in_min: int = field(
        metadata=field_options(alias="overallAverageTravelTimeInMin")
    )
    overall_mileage_in_km: int = field(metadata=field_options(alias="overallMileageInKm"))
    overall_travel_time_in_min: int = field(metadata=field_options(alias="overallTravelTimeInMin"))
    vehicle_type: VehicleType = field(metadata=field_options(alias="vehicleType"))
    detailed_statistics: list[StatisticsEntry] = field(
        metadata=field_options(alias="detailedStatistics")
    )
